using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using MacGruber;
using MVR.FileManagementSecure;

namespace CheesyFX
{
    public partial class LightMeUp
    {
        public JSONStorableFloat blendPresets = new JSONStorableFloat("Blend presets 0/1", 0f,0f,1f);
        

        public JSONStorableDirectory lightShowUrl;
        public JSONStorableBool lightshowEnabled = new JSONStorableBool("Lightshow enabled", false);
        public JSONStorableBool allowMirroredLR = new JSONStorableBool("Allow auto mirror L/R", true);
        public JSONStorableBool allowMirroredB = new JSONStorableBool("Allow auto mirror backlight", true);
        public JSONStorableFloat blendSpeedMean = new JSONStorableFloat("BlendSpeedMean", .2f,.01f,10f, false);
        public JSONStorableFloat blendSpeedDelta = new JSONStorableFloat("BlendSpeedDelta", .1f,0f,10f, false);
        public JSONStorableFloat delayMean = new JSONStorableFloat("DelayMean", 0f,0f,10f, false);
        public JSONStorableFloat delayDelta = new JSONStorableFloat("DelayDelta", 0f,0f,10f, false);
        
        private JSONStorableStringChooser presetChooser0 = new JSONStorableStringChooser("presetChooser0",
            null, "", "Preset0");
        private JSONStorableStringChooser presetChooser1 = new JSONStorableStringChooser("presetChooser1",
            null, "", "Preset1");

        private JSONStorableString lightshowInfoLeft = new JSONStorableString("lightshowDirectory", "");
        private JSONStorableString lightshowInfoRight = new JSONStorableString("lightshowFiles", "");
        
        private List<string> presetChoices;
        private Dictionary<int, string> lightShowPresets = new Dictionary<int, string>();
        private int currentPresetId;
        private bool isBlending { get; set; }
        private float blendVal { get; set; }
        private float blendSpeed { get; set; }
        private float blendTimer { get; set; }

        private void Register()
        {
            UpdatePresetChoices();
            presetChooser0.setCallbackFunction += SetBlendPreset0;
            presetChooser1.setCallbackFunction += SetBlendPreset1;
            
            lightShowUrl = new JSONStorableDirectory("LightShowDirectory", saveDir + "/Presets", SetLightShowDir, saveDir);
            SetLightShowDir(lightShowUrl.defaultVal);
            
            RegisterStringChooser(presetChooser0);
            RegisterStringChooser(presetChooser1);
            RegisterFloat(blendPresets);
            
            RegisterUrl(lightShowUrl);
            RegisterBool(allowMirroredLR);
            RegisterBool(allowMirroredB);
            RegisterBool(lightshowEnabled);
            RegisterFloat(blendSpeedMean);
            RegisterFloat(blendSpeedDelta);
            RegisterFloat(delayMean);
            RegisterFloat(delayDelta);
            
        }
        
        private void SetColor0(JSONStorableColor colorJson)
        {
            headLights[0].color.val = colorJson.val;
        }
        
        private void SetBlendPreset0(string name)
        {
            if(String.IsNullOrEmpty(presetChooser0.val) || !presetChoices.Contains(name)) return;
            string path = saveDir + "/Presets/" + name;
            var preset = LoadJSON(path).AsObject;
            headLights[0].blendPreset0 = new BlendPreset(preset["Left"].AsObject);
            headLights[1].blendPreset0 = new BlendPreset(preset["Center"].AsObject);
            headLights[2].blendPreset0 = new BlendPreset(preset["Right"].AsObject);
            headLights[3].blendPreset0 = new BlendPreset(preset["Back"].AsObject);
        }
        
        private void SetBlendPreset1(string name)
        {
            if(String.IsNullOrEmpty(presetChooser1.val) || !presetChoices.Contains(name)) return;
            string path = saveDir + "/Presets/" + name;
            var preset = LoadJSON(path).AsObject;
            headLights[0].blendPreset1 = new BlendPreset(preset["Left"].AsObject);
            headLights[1].blendPreset1 = new BlendPreset(preset["Center"].AsObject);
            headLights[2].blendPreset1 = new BlendPreset(preset["Right"].AsObject);
            headLights[3].blendPreset1 = new BlendPreset(preset["Back"].AsObject);
        }

        private bool currentMirrorLR;
        private bool currentMirrorB;
        private void InitBlendToNext()
        {
            if (lightShowPresets.Count == 0)
            {
                "Lightshow directory not set or it has zero presets in it.".Print();
                lightshowEnabled.val = false;
                return;
            }
            System.Random rnd = new System.Random();
            int id = rnd.Next(lightShowPresets.Count);
            bool mirroredLR = false;
            bool mirroredB = false;
            if (id == currentPresetId)
            {
                if(allowMirroredLR.val && allowMirroredB.val)
                {
                    mirroredLR = rnd.Next(2) == 1;
                    if (mirroredLR == currentMirrorLR) mirroredB = !currentMirrorB;
                    else mirroredB = rnd.Next(2) == 1;
                }
                else if (allowMirroredLR.val) mirroredLR = !currentMirrorLR;
                else if (allowMirroredB.val) mirroredB = !currentMirrorB;
                else
                {
                    for (int i = 0; i < 10; i++)
                    {
                        id = rnd.Next(lightShowPresets.Count);
                        if(id != currentPresetId) break;
                    }
                }
            }
            else
            {
                if(allowMirroredLR.val) mirroredLR = rnd.Next(2) == 1;
                if(allowMirroredB.val) mirroredB = rnd.Next(2) == 1;
            }
            // bool mirroredLR = id == currentPreset || (allowMirroredLR.val && rnd.Next(2) == 1);
            // bool mirroredB = id == currentPreset || (allowMirroredB.val && rnd.Next(2) == 1);
            currentPresetId = id;
            currentMirrorLR = mirroredLR;
            currentMirrorB = mirroredB;
            var url = lightShowPresets[id];
            // $"{url} {mirroredLR} {mirroredB}".Print();
            JSONClass next = LoadJSON(url).AsObject;
            JSONClass currentSettings = GetLightSettingsJSON();
            
            headLights[0].current = new BlendPreset(currentSettings["Left"].AsObject);
            headLights[1].current = new BlendPreset(currentSettings["Center"].AsObject);
            headLights[2].current = new BlendPreset(currentSettings["Right"].AsObject);
            headLights[3].current = new BlendPreset(currentSettings["Back"].AsObject);
            
            headLights[0].next = new BlendPreset(next["Left"].AsObject, mirroredLR);
            headLights[1].next = new BlendPreset(next["Center"].AsObject);
            headLights[2].next = new BlendPreset(next["Right"].AsObject, mirroredLR);
            headLights[3].next = new BlendPreset(next["Back"].AsObject, mirroredB);

            isBlending = true;
            blendVal = 0f;
            blendSpeed = NormalDistribution.GetValue(blendSpeedMean.val, blendSpeedDelta.val, sharpness:2f);
            
        }
        
        private void BlendToNext(float val)
        {
            var b = editSymmetrical.val;
            if(b) editSymmetrical.val = false;
            for (int i = 0; i < 4; i++)
            {
                headLights[i].BlendToNext(val);
            }
            if(b) editSymmetrical.val = true;
        }
        
        private void BlendPresets(float val)
        {
            var b = editSymmetrical.val;
            if(b) editSymmetrical.val = false;
            for (int i = 0; i < 4; i++)
            {
                headLights[i].BlendPresets(val);
            }
            if(b) editSymmetrical.val = true;
        }
        
        private void UpdatePresetChoices()
        {
            var presets = FileManagerSecure.GetFiles(saveDir + "/Presets");
            presetChoices = presets.Select(x => x.Replace(saveDir + "/Presets\\", ""))
                .Where(x => x.EndsWith(".json")).ToList();
            presetChooser0.choices = presetChoices;
            presetChooser1.choices = presetChoices;
            SetBlendPreset0(presetChooser0.val);
            SetBlendPreset1(presetChooser1.val);
        }

        private void CreateLightshowUI()
        {
            RemoveToggleArray(toggleArray);
            Utils.RemoveUIElements(this, UIElements);
            Utils.RemoveUIElements(this, LightUIElements);
            UIElements.Clear();
            
            UIElements.Add(Utils.SetupButton(this, "Back", CreateUI, true));
            var info = Utils.SetupInfoOneLine(this, "Manual blending", false);
            info.height = 50f;
            info.text.fontSize = 30;
            info.text.alignment = TextAnchor.MiddleCenter;
            UIElements.Add(info);
            UIElements.Add(CreateFilterablePopup(presetChooser0));
            UIElements.Add(CreateFilterablePopup(presetChooser1,true));
            UIElements.Add(CreateSlider(blendPresets, false));
            var spacer = CreateSpacer(true);
            spacer.height = 120f;
            UIElements.Add(spacer);
            info = Utils.SetupInfoOneLine(this, "Lightshow", false);
            info.height = 50f;
            info.text.fontSize = 30;
            info.text.alignment = TextAnchor.MiddleCenter;
            UIElements.Add(info);
            // spacer = CreateSpacer(true);
            // spacer.height = 50f;
            // UIElements.Add(spacer);
            UIElements.Add(CreateToggle(allowMirroredLR, true));
            UIElements.Add(CreateToggle(allowMirroredB, true));
            var button = CreateButton("Select lightshow directory",false);
            lightShowUrl.RegisterFileBrowseButton(button.button);
            UIElements.Add(button);
            UIElements.Add(CreateToggle(lightshowEnabled));
            UIElements.Add(Utils.SetupButton(this, "Trigger blend to random", InitBlendToNext, true));
            UIElements.Add(CreateSlider(blendSpeedMean, false));
            UIElements.Add(CreateSlider(blendSpeedDelta, false));
            UIElements.Add(CreateSlider(delayMean, true));
            UIElements.Add(CreateSlider(delayDelta, true));
            var textfied = CreateTextField(lightshowInfoLeft);
            textfied.height = 400f;
            UIElements.Add(textfied);
            textfied = CreateTextField(lightshowInfoRight,true);
            textfied.height = 400f;
            UIElements.Add(textfied);
        }
    }
    
    public class BlendPreset
    {
        public bool enabled;
        public Vector3 hsv;
        public float intensity;
        public float spotAngle;
        public float range;
        public float shadowStrength;
        public float offsetX;
        public float offsetY;
        public float offsetZ;

        public BlendPreset(JSONClass jc, bool mirrored = false, bool mirroredBacklight = false)
        {
            enabled = jc["Enabled"].AsBool;
            intensity = jc["Base intensity"].AsFloat;
            var col = jc["Color"].AsArray;
            hsv = new Vector3(col[0].AsFloat, col[1].AsFloat, col[2].AsFloat);
            spotAngle = jc["Spot Angle"].AsFloat;
            range = jc["Range"].AsFloat;
            shadowStrength = jc["Shadow Strength"].AsFloat;
            offsetX = jc["Horizontal Offset"].AsFloat;
            if (mirrored) offsetX *= -1f;
            offsetY = jc["Vertical Offset"].AsFloat;
            if (jc.HasKey("Depth offset from player")) offsetZ = jc["Depth offset from player"].AsFloat;
            else offsetZ = jc["Depth behind the target"].AsFloat;
        }

        public BlendPreset Mirror()
        {
            offsetX *= -1;
            return this;
        }
    }
}