using System;
using System.Linq;
using System.Collections.Generic;
using MacGruber;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using UnityEngine;
using UnityEngine.UI;


namespace CheesyFX
{
    public partial class LightMeUp : MVRScript
    {
        public float dist => viewScan.dist.val;
        public Vector3 targetPoint => viewScan.targetPoint.val;
        private ViewScan viewScan;
        private static bool _toggleState = true;
        public static List<HeadLight> headLights = new List<HeadLight>();
        public static Transform cameraTransform;
        
        private static JSONStorableStringChooser lightChooser;
        public static JSONStorableBool editSymmetrical;
        private static JSONStorableBool autoFocus = new JSONStorableBool("Auto focus", true);
        private static JSONStorableBool autoIntensity = new JSONStorableBool("Auto intensity", true);
        private static JSONStorableBool debug = new JSONStorableBool("Debug", false);
        
        private static JSONStorableStringChooser leftKeyChooser;
        private static JSONStorableStringChooser centerKeyChooser;
        private static JSONStorableStringChooser rightKeyChooser;
        private static JSONStorableStringChooser backKeyChooser;
        private static JSONStorableStringChooser allKeyChooser;
        private static JSONStorableStringChooser mod0Chooser;
        private static JSONStorableStringChooser mod1Chooser;
        private static JSONStorableStringChooser mod2Chooser;

        private JSONStorableUrl loadURL;
        private const string saveDir = "Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp";
        
        public static JSONStorableBool useKeybindings;
        

        private bool sideLightEnabled => headLights[0].enabled.val || headLights[2].enabled.val;
        
        private static RectTransform versionRectTransform;
        private static RectTransform buttonsTransform;
        private static RectTransform toggleButtonTransform;
        
        private ColorBlock activeColorBlock;
        public static ColorBlock disabledColorBlock;


        private static KeyCode[] toggleKeys = {
            KeyCode.F9,
            KeyCode.F10,
            KeyCode.F11,
            KeyCode.F12,
            KeyCode.KeypadPlus,
        };
        private static KeyCode[] modifiers = {
            KeyCode.RightControl,
            KeyCode.RightShift,
            KeyCode.LeftAlt
        };
        public static bool disableSymmetry;

        private static JSONStorableFloat IntensityRate = new JSONStorableFloat("Intensity Rate", .002f, .001f, .05f);

        private List<object> UIElements = new List<object>();
        private List<object> LightUIElements = new List<object>();

        public static Color[] defaultColors = 
        {
            new Color(1f, 0.89f, 0.78f),
            new Color(1f,1f,1f),
            new Color(0.78f, 0.98f, 1f),
            new Color(0.93f, 1f, 0.71f)
        };
        
        private static Dictionary<Atom, bool> sceneLights = new Dictionary<Atom, bool>();
        private static bool isSceneLit = true;
        public static bool isOnVR;
        private static bool keybindingsPresent;

        private UIDynamicTextInfo lightInfo;

        private JSONClass sceneLightConfig;
        private MVRScript sessionInstance;
        private bool abort;

        private JSONStorable scenePlugin;
        private string scenePluginUid;

        private void OnEnable()
        {
            abort = transform.parent.parent.name != "SessionPluginManager";
            if (abort)
            {
                SuperController.LogError("FocusOnMe!: This cannot be installed as a scene or atom plugin. Use it as a SESSION plugin.");
            }
        }

        public override void Init()
        {
            if (abort) return;
            FileManagerSecure.CreateDirectory(saveDir+"/Presets");
            // FileManagerSecure.CreateDirectory(saveDir+"/Scenes");
            loadURL = new JSONStorableUrl("loadURL", "", UILoadLightSettingsFromJSON, "json", true);
            loadURL.hideExtension = true;
            loadURL.setCallbackFunction -= UILoadLightSettingsFromJSON;
            loadURL.allowFullComputerBrowse = false;
            loadURL.allowBrowseAboveSuggestedPath = true;
            loadURL.SetFilePath(saveDir+"/Presets/");
            loadURL.setCallbackFunction += UILoadLightSettingsFromJSON;
            Utils.OnInitUI(CreateUIElement);
            SetCameraRig();
            disabledColorBlock = ColorBlock.defaultColorBlock;
            disabledColorBlock.normalColor *= .5f;
            disabledColorBlock.highlightedColor *= .5f;
            disabledColorBlock.pressedColor *= .5f;
            headLights = new List<HeadLight>()
            {
                new HeadLight(this, 0),
                new HeadLight(this, 1),
                new HeadLight(this, 2),
                new HeadLight(this, 3)
            };
            
            CreateMainHUD();
            
            lightChooser = new JSONStorableStringChooser(
                "LightChooser",
                headLights.Select(x => x.name).ToList(),
                "Center",
                "Light",
                CreateLightUI
                );
            List<string> keyOptions = Enum.GetNames(typeof(KeyCode)).ToList();
            leftKeyChooser = new JSONStorableStringChooser("Key0", keyOptions,
                toggleKeys[0].ToString(), "Key Left", SetKey);
            centerKeyChooser = new JSONStorableStringChooser("Key1", keyOptions,
                toggleKeys[1].ToString(), "Key Center", SetKey);
            rightKeyChooser = new JSONStorableStringChooser("Key2", keyOptions,
                toggleKeys[2].ToString(), "Key Right", SetKey);
            backKeyChooser = new JSONStorableStringChooser("Key3", keyOptions,
                toggleKeys[3].ToString(), "Key Back", SetKey); 
            allKeyChooser = new JSONStorableStringChooser("Key4", keyOptions,
                toggleKeys[4].ToString(), "Key All", SetKey);
            
            mod0Chooser = new JSONStorableStringChooser("Mod0", keyOptions,
                modifiers[0].ToString(), "Modifier0", SetModifier);
            mod1Chooser = new JSONStorableStringChooser("Mod1", keyOptions,
                modifiers[1].ToString(), "Modifier1", SetModifier);
            mod2Chooser = new JSONStorableStringChooser("Mod2", keyOptions,
                modifiers[2].ToString(), "Modifier2", SetModifier);

            useKeybindings = new JSONStorableBool("Use AcidBubbles Keybindings", false, UseKeybindingsCallback);
            editSymmetrical = new JSONStorableBool("Edit Sym", false);

            autoFocus.setCallbackFunction += OnAutoFocusChanged;
            autoIntensity.setCallbackFunction += OnAutoIntensityChanged;
            debug.setCallbackFunction += ToggleDebugLines;

            blendPresets.setCallbackFunction += BlendPresets;
            
            SetupKeyInput();
            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);
            CreateUI();
            SuperController.singleton.onBeforeSceneSaveHandlers += ParseSceneLightConfig;
            SuperController.singleton.onSceneLoadedHandlers += GetOrAddScenePlugin;
            SuperController.singleton.onSceneLoadedHandlers += SetSceneLit;
            Register();
        }

        public void Start()
        {
            if (abort)
            {
                enabledJSON.val = false;
                return;
            }
            viewScan = Utils.FindWithinSamePlugin<ViewScan>(this);
            LoadDefaultLightSettings();
            LoadInteractionSettings();
            // GetOrAddScenePlugin();
            Toggle();
            // viewScan.uses.val.ToString().Print();
        }

        public void OnBindingsListRequested(List<object> bindings)
        {
            bindings.Add(new Dictionary<string, string>
            {
                { "CheesyFX", "LightMeUp" }
            });

            for (int i = 0; i < toggleKeys.Length; i++)
            {
                bindings.Add(KeyInput.keysHeldDown[i]);
            }

            bindings.Add(new JSONStorableAction("All Reset Intensity", () => ResetIntensity(4, true)));
            bindings.Add(new JSONStorableAction("Left Reset Intensity", () => ResetIntensity(0, true)));
            bindings.Add(new JSONStorableAction("Center Reset Intensity", () => ResetIntensity(1, true)));
            bindings.Add(new JSONStorableAction("Right Reset Intensity", () => ResetIntensity(2, true)));
            bindings.Add(new JSONStorableAction("Back Reset Intensity", () => ResetIntensity(3, true)));
            
            bindings.Add(new JSONStorableAction("All Toggle Shadows", () => ToggleShadows(4, true)));
            bindings.Add(new JSONStorableAction("Left Toggle Shadows", () => ToggleShadows(0, true)));
            bindings.Add(new JSONStorableAction("Center Toggle Shadows", () => ToggleShadows(1, true)));
            bindings.Add(new JSONStorableAction("Right Toggle Shadows", () => ToggleShadows(2, true)));
            bindings.Add(new JSONStorableAction("Back Toggle Shadows", () => ToggleShadows(3, true)));
            
            bindings.Add(new JSONStorableAction("Toggle Scene Lights", () => ToggleSceneLighting(true)));
            bindings.Add(new JSONStorableAction("Swap L/R colors", () => SwapLeftRightColor(true)));
            
            bindings.Add(new JSONStorableAction("Left Intensity + 0.1", () => headLights[0].intensity.val+=.1f));
            bindings.Add(new JSONStorableAction("Center Intensity + 0.1", () => headLights[1].intensity.val+=.1f));
            bindings.Add(new JSONStorableAction("Right Intensity + 0.1", () => headLights[2].intensity.val+=.1f));
            bindings.Add(new JSONStorableAction("Back Intensity + 0.1", () => headLights[3].intensity.val+=.1f));
            bindings.Add(new JSONStorableAction("All Intensity + 0.1", () => headLights.ForEach(x=>x.intensity.val+=.1f)));
            
            bindings.Add(new JSONStorableAction("Left Intensity - 0.1", () => headLights[0].intensity.val-=.1f));
            bindings.Add(new JSONStorableAction("Center Intensity - 0.1", () => headLights[1].intensity.val-=.1f));
            bindings.Add(new JSONStorableAction("Right Intensity - 0.1", () => headLights[2].intensity.val-=.1f));
            bindings.Add(new JSONStorableAction("Back Intensity - 0.1", () => headLights[3].intensity.val-=.1f));
            bindings.Add(new JSONStorableAction("All Intensity - 0.1", () => headLights.ForEach(x=>x.intensity.val-=.1f)));

            keybindingsPresent = true;
        }
        
        private void Toggle(int id = 4)
        {
            if (!enabled && !_toggleState) return;
            if(id == 4 || !_toggleState)
            {
                toggleButtonTransform.gameObject.SetActive(_toggleState);
                _toggleState = !_toggleState;
                headLights.ForEach(x => x.EnabledCallback(_toggleState && x.enabled.val));
                // if (viewScan == null) return;
                if (!_toggleState) {
                    viewScan.uses.val--;
                }
                else
                {
                    viewScan.uses.val++;
                    GetOrAddScenePlugin();
                }
            }
            else
            {
                headLights[id].Toggle();
                if (_toggleState) return;
                toggleButtonTransform.gameObject.SetActive(false);
                _toggleState = true;
                GetOrAddScenePlugin();
            }
        }

        private void GetOrAddScenePlugin()
        {
            var core = SuperController.singleton.GetAtomByUid("CoreControl");
            var uid = core.GetStorableIDs().FirstOrDefault(x => x.EndsWith("CheesyFX.FocusOnMe_ScenePlugin"));
            if(uid == null)
            {
                var preset = CreateFakePluginPresetJSON();
                var presetManagerControl = core.GetStorableByID("PluginManagerPresets");
                var presetManager = presetManagerControl.GetComponentInChildren<PresetManager>();
                presetManager.LoadPresetFromJSON(preset, true);
                uid = core.GetStorableIDs().FirstOrDefault(x => x.EndsWith("CheesyFX.FocusOnMe_ScenePlugin"));
            }
            scenePlugin = core.GetStorableByID(uid);
            scenePlugin.SendMessage("DeferredFindSessionPlugin");
            scenePluginUid = uid.Substring(0,uid.IndexOf("_"));
        }

        private void RemoveScenePlugin()
        {
            var core = SuperController.singleton.GetAtomByUid("CoreControl");
            var uid = core.GetStorableIDs().FirstOrDefault(x => x.EndsWith("CheesyFX.FocusOnMe_ScenePlugin"));
            if (uid == null) return;
            string id = uid.Substring(0, uid.IndexOf('_'));
            var pm = core.GetComponentsInChildren<MVRPluginManager>().First(x => x.name == "ScenePluginManager");
            pm.RemovePluginWithUID(id);
        }

        private void ParseSceneLightConfig()
        {
            if (!enabled || !_toggleState)
            {
                RemoveScenePlugin();
                return;
            }
            
            GetOrAddScenePlugin();
            scenePlugin.gameObject.SendMessage("FetchSceneLightConfig", GetLightSettingsJSON());
            scenePlugin.gameObject.SendMessage("FetchLightshowConfig", GetLightshowJSON());
        }

        private void OnAutoFocusChanged(bool val)
        {
            // if (!autoIntensity.val) viewScan.uses.val += val? 1 : -1;
            foreach (var headLight in headLights) headLight.rotation.SetInteractble(!val);
        }

        private void OnAutoIntensityChanged(bool val)
        {
            // if (!autoFocus.val) viewScan.uses.val += val? 1 : -1;
            if(!val) headLights.ForEach(x => x.light.intensity = x.intensity.val);
        }

        private void PrintHelp()
        {
            string help = "LightMeUp shortcuts:\n\n" +
                          "Main Keys:\n" +
                          $"Left:\t\t\t\t\t\t\t\t\t\t\t{toggleKeys[0]}\n" +
                          $"Center:\t\t\t\t\t\t\t\t\t\t{toggleKeys[1]}\n" +
                          $"Right:\t\t\t\t\t\t\t\t\t\t{toggleKeys[2]}\n" +
                          $"All:\t\t\t\t\t\t\t\t\t\t\t{toggleKeys[3]}\n\n" +
                          "Tap & Hold:\t\t\t\t\t\t\t\tincrease intensity\n" +
                          "Double Tap & Hold:\t\t\t\t\tdecrease intensity\n\n" +
                          "Modifiers:\n" +
                          $"Reset Intensity:\t\t\t\t\t\t{modifiers[0]}+<any>\n" +
                          $"Toggle Shadows:\t\t\t\t\t\t{modifiers[1]}+<any>\n" +
                          $"SceneLights / Swap color:\t\t{modifiers[2]}+<any>\n";
            help.Print();
        }

        private void SetupKeyInput()
        {
            KeyInput.onSingle = OnSingleTap;
            KeyInput.onDuringSingleStart = key => disableSymmetry = true;
            KeyInput.onDuringSingleEnd = key => disableSymmetry = false;
            KeyInput.onDuringDoubleStart = key => disableSymmetry = true;
            KeyInput.onDuringDoubleEnd = key => disableSymmetry = false;
            KeyInput.duringSingleDown = key => AdjustIntensity(key, 1f);
            KeyInput.duringDoubleDown = key => AdjustIntensity(key, -1f);

            KeyInput.SetUpKeys(new []
            {
                "Left (Toggle & Intensity)",
                "Center (Toggle & Intensity)",
                "Right (Toggle & Intensity)",
                "Back (Toggle & Intensity)",
                "All (Toggle & Intensity)"
            });
            KeyInput.modifiers = modifiers;
        }

        private void SetKey(JSONStorableStringChooser jString)
        {
            int id = int.Parse(jString.name[3].ToString());
            toggleKeys[id] = (KeyCode)Enum.Parse(typeof(KeyCode), jString.val);
        }
        
        private void SetModifier(JSONStorableStringChooser jString)
        {
            int id = int.Parse(jString.name[3].ToString());
            KeyInput.modifiers[id] = (KeyCode)Enum.Parse(typeof(KeyCode), jString.val);
        }

        private void OnSingleTap(int keyId, KeyCode modifier)
        {
            if(modifier == KeyCode.LeftAlt)
            {
                if(keyId==4) ToggleSceneLighting();
                else SwapLeftRightColor();
            }
            else
            {
                if(modifier == KeyCode.LeftControl || modifier == KeyCode.RightControl) ResetIntensity(keyId);
                else if (modifier == KeyCode.LeftShift || modifier == KeyCode.RightShift) ToggleShadows(keyId);
                else Toggle(keyId);
            }
        }

        private void UseKeybindingsCallback(bool b)
        {
            if (b && !keybindingsPresent)
            {
                SuperController.LogError("LightMeUp: Plugin AcidBubbles.Keybindings not found.\n" +
                                         "Add Keybindings as a session plugin first, then reload this plugin.");
                useKeybindings.valNoCallback = false;
            }
        }

        private void SetSceneLit()
        {
            isSceneLit = true;
        }

        private void SetCameraRig()
        {
            if (SuperController.singleton.isOpenVR)
            {
                cameraTransform = SuperController.singleton.ViveCenterCamera.transform;
                isOnVR = true;
            }
            else if (SuperController.singleton.isOVR)
            {
                cameraTransform = SuperController.singleton.OVRCenterCamera.transform;
                isOnVR = true;
            }
            else if (SuperController.singleton.IsMonitorOnly)
            {
                cameraTransform = SuperController.singleton.MonitorCenterCamera.transform;
            }
        }

        private void CreateLightUI(JSONStorableStringChooser jString)
        {
            // lightInfo.text.text = jString.val + " Light Settings";
            Utils.RemoveUIElements(this, LightUIElements);
            LightUIElements = headLights[lightChooser.choices.IndexOf(jString.val)].CreateUI();
        }


        public void Update()
        {
            if (!useKeybindings.val)
            {
                for (int i = 0; i < toggleKeys.Length; i++)
                {
                    if (Input.GetKeyDown(toggleKeys[i])) KeyInput.keysHeldDown[i].val = true;
                    else if (Input.GetKeyUp(toggleKeys[i])) KeyInput.keysHeldDown[i].val = false;
                }
            }

            if (!_toggleState) return;

            // Vector3 focusPoint = cameraTransform.position + cameraTransform.forward * dist;
            

            if (isBlending)
            {
                blendVal += Time.deltaTime * blendSpeed;
                BlendToNext(blendVal);
                if (blendVal > 1f)
                {
                    isBlending = false;
                }
            }
            else if(lightshowEnabled.val)
            {
                blendTimer -= Time.deltaTime;
                if (blendTimer < 0)
                {
                    blendTimer = NormalDistribution.GetValue(delayMean.val, delayDelta.val, 2f);
                    InitBlendToNext();
                }
                
            }
            
            if (autoFocus.val)
            {
                if (headLights[0].enabled.val) headLights[0].UpdateRotation(targetPoint);
                if (headLights[2].enabled.val) headLights[2].UpdateRotation(targetPoint);
                if (headLights[3].enabled.val)
                {
                    var lPos = headLights[3].lightObject.transform.localPosition;
                    lPos.z = headLights[3].offsetZ.val + dist;
                    headLights[3].lightObject.transform.localPosition = lPos;
                    headLights[3].UpdateRotation(targetPoint);
                }

                if (debug.val)
                {
                    for (int i = 0; i < debugLines.Count; i++)
                    {
                        var lr = debugLines[i].GetComponent<LineRenderer>();
                        lr.SetPosition(1, debugLines[i].transform.InverseTransformPoint(targetPoint));
                    }
                }
            }

            if (autoIntensity.val)
            {
                float distFactor = 1f + dist * .2f;
                for (int i = 0; i < headLights.Count; i++)
                {
                    headLights[i].light.intensity = headLights[i].intensity.val * distFactor;
                }
            }
        }

        private static void AdjustIntensity(int keyId, float direction)
        {
            float increment = IntensityRate.val * direction;
            if (keyId < 3) headLights[keyId].intensity.val += increment;
            else headLights.ForEach(x => x.intensity.val += increment);
        }

        private static void ResetIntensity(int id, bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            if(id == 3) headLights.ForEach(x => x.intensity.SetValToDefault());
            else
            {
                bool sym = editSymmetrical.val;
                editSymmetrical.val = false;
                headLights[id].intensity.SetValToDefault();
                editSymmetrical.val = sym;
            }
        }

        private static void SwapLeftRightColor(bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            HSVColor leftColor = headLights[0].color.val;
            headLights[0].color.val = headLights[2].color.val;
            headLights[2].color.val = leftColor;
        }

        private static void ToggleShadows(int id, bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            if (id == 3)
            {
                headLights.ForEach(x => x.castShadows.val = !x.castShadows.val);
            }
            else headLights[id].castShadows.val = !headLights[id].castShadows.val;
        }

        private static void ToggleSceneLighting(bool fromKeybindings = false)
        {
            if (fromKeybindings && !useKeybindings.val) return;
            if(isSceneLit) sceneLights.Clear();
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "InvisibleLight")
                {
                    JSONStorableBool on = atom.GetStorableByID("Light").GetBoolJSONParam("on");
                    if (isSceneLit)
                    {
                        sceneLights.Add(atom, on.val);
                        on.val = false;
                    }
                    else
                    {
                        if (sceneLights.ContainsKey(atom)) on.val = sceneLights[atom];
                        else sceneLights.Add(atom, on.val);
                    }
                }
            }
            isSceneLit = !isSceneLit;
        }

        public void OnDisable()
        {
            // headLights.ForEach(x => x.enabled.val = false);
            if(_toggleState) Toggle(4);
            if(!isSceneLit) ToggleSceneLighting();
        }
        
        public void OnDestroy()
        {
            headLights.ForEach(x => x.Destroy());
            versionRectTransform.anchorMax = new Vector2(.37f, 0f);
            Destroy(toggleButtonTransform.gameObject);
            if(!isSceneLit) ToggleSceneLighting();
            SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            SuperController.singleton.onBeforeSceneSaveHandlers -= ParseSceneLightConfig;
            SuperController.singleton.onSceneLoadedHandlers -= GetOrAddScenePlugin;
            SuperController.singleton.onSceneLoadedHandlers -= SetSceneLit;
            debugLines.ForEach(Destroy);
            var cc = SuperController.singleton.GetAtomByUid("CoreControl");
            var pm = cc.GetComponentsInChildren<MVRPluginManager>()[1];
            pm.RemovePluginWithUID(scenePluginUid);
        }

        private UIDynamicToggleArray toggleArray;
        private void CreateUI()
        {
            Utils.RemoveUIElements(this, UIElements);
            UIElements.Clear();
            
            toggleArray = CreateToggleArray(new JSONStorableBool[] { autoFocus, autoIntensity, editSymmetrical, debug });
            // UIElements.Add();
            // UIElements.Add(CreateToggle(autoFocus));
            // UIElements.Add(CreateToggle(autoIntensity, true));
            var twinButton = Utils.SetupTwinButton(this, "Save preset", UIGetLightSettingsJSONDialog, "Load preset", delegate {  }, false);
            loadURL.RegisterFileBrowseButton(twinButton.buttonRight);
            UIElements.Add(twinButton);
            
            UIElements.Add(Utils.SetupTwinButton(this, "Load scene config", ReloadSceneConfig, "Reset lights", ResetLights, true));
            
            UIElements.Add(Utils.SetupTwinButton(this, "Swap L/R colors", () => SwapLeftRightColor(), "ToggleSceneLights", () => ToggleSceneLighting(), true));
            UIElements.Add(Utils.SetupTwinButton(this, "Save as default", SaveDefaultLightSettings, "Edit lightshow", CreateLightshowUI, false));
            UIElements.Add(CreateToggle(lightshowEnabled));
            // lightInfo = Utils.SetupInfoOneLine(this, "Center Light", false);
            // lightInfo.height = 50f;
            // lightInfo.text.fontSize = 32;
            // lightInfo.text.alignment = TextAnchor.MiddleCenter;
            // UIElements.Add(lightInfo);
            UIElements.Add(CreateToggle(editSymmetrical, false));
            UIElements.Add(Utils.SetupTwinButton(this, "Input Options", CreateInputUI, "Print Help", PrintHelp, true));
            UIElements.Add(CreatePopup(lightChooser));
            
            CreateLightUI(lightChooser);
            
        }

        private void CreateInputUI()
        {
            RemoveToggleArray(toggleArray);
            Utils.RemoveUIElements(this, UIElements);
            Utils.RemoveUIElements(this, LightUIElements);
            UIElements.Clear();
            UIElements.Add(Utils.SetupTwinButton(this, "Save input config", SaveInputSettings, "Reset input", ResetInput, true));
            UIElements.Add(Utils.SetupButton(this, "Back", CreateUI, false));
            
            UIElements.Add(CreateSlider(KeyInput.doubleTapTimeout, false));
            UIDynamicSlider slider = CreateSlider(IntensityRate, true);
            slider.valueFormat = "{0.000}";
            UIElements.Add(slider);
            UIElements.Add(CreateSlider(IntensityRate, true));
            
            UIElements.Add(CreateFilterablePopup(allKeyChooser));
            UIElements.Add(CreateFilterablePopup(leftKeyChooser));
            UIElements.Add(CreateFilterablePopup(centerKeyChooser));
            UIElements.Add(CreateFilterablePopup(rightKeyChooser));
            UIElements.Add(CreateFilterablePopup(backKeyChooser));

            UIElements.Add(CreateFilterablePopup(mod0Chooser, true));
            UIElements.Add(CreateFilterablePopup(mod1Chooser, true));
            UIElements.Add(CreateFilterablePopup(mod2Chooser, true));
            UIElements.Add(CreateToggle(useKeybindings, false));
        }
        
        private void CreateMainHUD()
        {
            //.37f
            RectTransform panel = SuperController.singleton.GetAtomByUid("CoreControl").gameObject.transform.Find("WorldScaleAdjust/HUD/LowerHUDPivot/LowerHUDFlip/Scene Control Canvas/Panel").GetComponent<RectTransform>();
            versionRectTransform = panel.Find("VersionText").GetComponent<RectTransform>();
            versionRectTransform.anchorMax = new Vector2(.25f, 0f);
            headLights.ForEach(x => x.CreateHUDToggle(panel));
            
            toggleButtonTransform = manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
            toggleButtonTransform = Instantiate(toggleButtonTransform, panel);
            toggleButtonTransform.name = $"ToggleButton";
            toggleButtonTransform.anchorMax = new Vector2(.38f, 1);
            toggleButtonTransform.anchorMin = new Vector2(.25f, 0);
            toggleButtonTransform.offsetMax = new Vector2(0, -5f);
            toggleButtonTransform.offsetMin = new Vector2(0, 5f);
            Button button = toggleButtonTransform.GetComponent<Button>();
            Text buttonText = toggleButtonTransform.Find("Text").GetComponent<Text>();
            buttonText.text = "LightMeUp";
            button.onClick.AddListener(() => Toggle());
            toggleButtonTransform.gameObject.SetActive(false);
        }

        private static List<GameObject> debugLines = new List<GameObject>();
        private static void CreateDebugLine(int id)
        {
            var headLight = headLights[id];
            if (!headLight.enabled.val) return;
            Transform cameraTransform = headLights[id].lightObject.transform;
            GameObject myLine = new GameObject();
            myLine.transform.position = cameraTransform.position;
            myLine.transform.parent = cameraTransform;
            myLine.transform.rotation = new Quaternion();
            LineRenderer lr = myLine.AddComponent<LineRenderer>();
            lr.useWorldSpace = false;
            // Material whiteDiffuseMat = new Material(Shader.Find("Unlit/Texture"));
            // lr.material.shader = Shader.Find("Sprites/Default");
            lr.material = new Material(Shader.Find("Sprites/Default"));
            Color color = headLights[id].color.val.ToRGB();
            lr.SetColors(color, color);
            lr.SetWidth(0.001f, 0.001f);
            lr.SetPosition(0, new Vector3());
            lr.SetPosition(1, new Vector3(0f,0f,10f));
            debugLines.Add(myLine);
        }
        
        private void ToggleDebugLines(bool enabled)
        {
            if(enabled){
                for (int i = 0; i < 4; i++) CreateDebugLine(i);
            }
            else debugLines.ForEach(Destroy);
        }
        
        public UIDynamicToggleArray CreateToggleArray(JSONStorableBool[] bjsons, float width = 1f, int columns = -1)
        {
            if (columns == -1) columns = bjsons.Length;
            float rowCount = Mathf.Ceil((float)bjsons.Length / columns);
            float totalHeight = 70f + 45f * (rowCount - 1);
            GameObject prefab = new GameObject("ToggleArray");
            var uid = prefab.AddComponent<UIDynamicToggleArray>();
            RectTransform rt = prefab.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0, 1);
            rt.anchorMin = new Vector2(0, 1);
            // rt.offsetMax = new Vector2(1064, 0);
            // rt.offsetMin = new Vector2(10, -60);
            LayoutElement le = prefab.AddComponent<LayoutElement>();
            le.flexibleWidth = 1;
            le.minHeight = totalHeight;
            le.minWidth = 1064;
            le.preferredHeight = totalHeight;
            le.preferredWidth = 1064;
            
            RectTransform backgroundTransform =
                manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
            backgroundTransform = Instantiate(backgroundTransform, prefab.transform);
            backgroundTransform.name = "Background";
            backgroundTransform.anchorMax = new Vector2(1, 1);
            backgroundTransform.anchorMin = new Vector2(0, 0);
            // rt.offsetMax = new Vector2(1064, 0);
            // rt.offsetMin = new Vector2(10, -45);
            
            float padding = 5.0f;
            float itemWidth, rowOffset;
            itemWidth = 1024f * width / columns - padding;
            
            float x = 15;
            for (int i = 0; i < bjsons.Length; ++i)
            {
                int rowNumber = i / columns;
                rowOffset = 45f * rowNumber;
                if (i % columns == 0) x = 15f;
                rt = Instantiate(
                    manager.configurableTogglePrefab.transform,
                    prefab.transform,
                    false) as RectTransform;
                rt.name = "Toggle";
                rt.anchorMax = new Vector2(0, 1);
                rt.anchorMin = new Vector2(0, 1);
                rt.offsetMax = new Vector2(x + itemWidth, -15 - rowOffset);
                rt.offsetMin = new Vector2(x, -55 - rowOffset);
                Toggle toggle = rt.GetComponent<Toggle>();
                uid.toggles.Add(toggle);
                Text buttonText = rt.Find("Label").GetComponent<Text>();
                buttonText.text = bjsons[i].name;
                x += itemWidth + padding;
            }
            
            Transform t = CreateUIElement(prefab.transform);
            uid = t.gameObject.GetComponent<UIDynamicToggleArray>();
            for (int i = 0; i < bjsons.Length; ++i)
            {
                bjsons[i].RegisterToggle(uid.toggles[i]);
            }
            UIDynamic spacer = CreateSpacer(true);
            spacer.height = totalHeight;
            // spacer.transform.SetParent(t);
            uid.spacer = spacer;
            Destroy(prefab);
            return uid;
        }

        private void RemoveToggleArray(UIDynamicToggleArray uid)
        {
            RemoveSpacer(uid.spacer);
            RemoveSpacer(uid);
        }

        private JSONStorable FindSessionPlugin()
        {
            var core = SuperController.singleton.GetAtomByUid("CoreControl");
            var storableId = core.GetStorableIDs().FirstOrDefault(x => x.Contains(name));
            return storableId == null ? null : core.GetStorableByID(storableId);
            
        }
    }
    

    public class UIDynamicToggleArray : UIDynamic
    {
        public UIDynamic spacer;
        public List<Toggle> toggles = new List<Toggle>();
    }
    
    public static class Ext
    {
        public static void Print(this float v)
        {
            SuperController.LogMessage(v.ToString());
        }
        
        public static void Print(this string v)
        {
            SuperController.LogMessage(v);
        }
        
        public static void Print(this bool v)
        {
            SuperController.LogMessage(v.ToString());
        }
        
        public static void NullCheck(this object v)
        {
            SuperController.LogMessage((v == null).ToString());
        }
    }
}