using System;
using System.Linq;
using MVR.FileManagementSecure;
using SimpleJSON;
using MacGruber;

namespace CheesyFX
{
    public partial class LightMeUp
    {
        // public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        // {
        //     JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
        //     if (!abort) jc["LightMeUp"] = GetLightSettingsJSON();
        //     else jc["LightMeUp"] = sceneLightConfig;
        //     needsStore = true;
        //     return jc;
        // }
        //
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            // base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            // // if (isSessionInstance) return;
            // if (restorePhysical && restoreAppearance) // StoreType.Full
            // {
            //     if (jc.HasKey("LightMeUp")){
            //         // "LateRestoreFromJSON".Print();
            //         sceneLightConfig = jc["LightMeUp"].AsObject;
            //         if(abort) sessionInstance.gameObject.SendMessage("LoadLightSettingsFromJSON", sceneLightConfig);
            //     }
            // }
        }


        JSONClass GetInputSettingsJSON(){
            JSONClass jc = new JSONClass();
            autoFocus.Store(jc);
            useKeybindings.Store(jc);
            KeyInput.doubleTapTimeout.Store(jc);
            IntensityRate.Store(jc);
            jc["Key left"] = leftKeyChooser.val;
            jc["Key center"] = centerKeyChooser.val;
            jc["Key right"] = rightKeyChooser.val;
            jc["Key back"] = backKeyChooser.val;
            jc["Mod0"] = mod0Chooser.val;
            jc["Mod1"] = mod1Chooser.val;
            jc["Mod2"] = mod2Chooser.val;
            return jc;
        }
        
        void LoadInputSettingsFromJSON(JSONClass jc)
        {
            autoFocus.Load(jc);
            useKeybindings.Load(jc);
            KeyInput.doubleTapTimeout.Load(jc);
            IntensityRate.Load(jc);
            leftKeyChooser.val = jc["Key left"];
            centerKeyChooser.val = jc["Key center"];
            rightKeyChooser.val = jc["Key right"];
            backKeyChooser.val = jc["Key back"];
            mod0Chooser.val = jc["Mod0"];
            mod1Chooser.val = jc["Mod1"];
            mod2Chooser.val = jc["Mod2"];
        }
        
        private void SaveInputSettings()
        {
            SaveJSON(GetInputSettingsJSON(), "Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp/InteractionSettings.json");
        }
        
        private void LoadInteractionSettings()
        {
            if (!FileManagerSecure.FileExists("Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp/InteractionSettings.json"))
                return;
            JSONClass jc = LoadJSON("Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp/InteractionSettings.json").AsObject;
            LoadInputSettingsFromJSON(jc);
        }

        private void ResetInput()
        {
            useKeybindings.SetValToDefault();
            KeyInput.doubleTapTimeout.SetValToDefault();
            IntensityRate.SetValToDefault();
            leftKeyChooser.SetValToDefault();
            centerKeyChooser.SetValToDefault();
            rightKeyChooser.SetValToDefault();
            backKeyChooser.SetValToDefault();
            mod0Chooser.SetValToDefault();
            mod1Chooser.SetValToDefault();
            mod2Chooser.SetValToDefault();
        }

        private JSONClass GetLightSettingsJSON(string name = null){
            JSONClass jc = new JSONClass();
            if (name != null) jc["Name"] = name;
            if (_toggleState) jc["_toggleState"].AsBool = _toggleState;
            foreach (var headLight in headLights)
            {
                jc[headLight.name] = headLight.GetJSON();
            }
            return jc;
        }
        
        private void LoadLightSettingsFromJSON(JSONClass jc)
        {
            if(jc.HasKey("_toggleState") && !_toggleState) Toggle();
            foreach (var headLight in headLights)
            {
                headLight.LoadJSON(jc[headLight.name].AsObject);
            }
        }
        
        public void UIGetLightSettingsJSON(string path)
        {
            if (string.IsNullOrEmpty(path)) return;
            string path1 = path.Replace('\\', '/');
            JSONClass jc = GetLightSettingsJSON(path1.Split('/').Last());
            SaveJSON(jc, path1+".json");
            SuperController.singleton.DoSaveScreenshot(path+".jpg");
            UpdatePresetChoices();
        }
        
        public void UIGetLightSettingsJSONDialog(){
            SuperController sc = SuperController.singleton;
            sc.GetMediaPathDialog(UIGetLightSettingsJSON, "json", saveDir+"/Presets", false, true, false, null, true, null, false, false);
            sc.mediaFileBrowserUI.SetTextEntry(true);
            if(sc.mediaFileBrowserUI.fileEntryField != null){
                string filename = DateTime.Now.ToString("yyyyMMdd_HH-mm-ss");;
                sc.mediaFileBrowserUI.fileEntryField.text = filename;
                sc.mediaFileBrowserUI.ActivateFileNameField();
            }
        }
        
        public void UILoadLightSettingsFromJSON(string url){
            JSONClass jc = LoadJSON(url).AsObject;
            if(jc != null) LoadLightSettingsFromJSON(jc);
        }

        private JSONClass GetLightshowJSON()
        {
            JSONClass jc = new JSONClass();
            presetChooser0.Store(jc);
            presetChooser1.Store(jc);
            lightshowEnabled.Store(jc);
            lightShowUrl.Store(jc);
            blendSpeedMean.Store(jc);
            blendSpeedDelta.Store(jc);
            delayMean.Store(jc);
            delayDelta.Store(jc);

            return jc;
        }

        private void LoadLightshowSettingsFromJSON(JSONClass jc)
        {
            presetChooser0.Load(jc);
            presetChooser1.Load(jc);
            lightshowEnabled.Load(jc);
            lightShowUrl.Load(jc);
            blendSpeedMean.Load(jc);
            blendSpeedDelta.Load(jc);
            delayMean.Load(jc);
            delayDelta.Load(jc);
        }
        
        public void SetLightShowDir(string url){
            lightShowPresets.Clear();
            var files = FileManagerSecure.GetFiles(url).Where(x => x.EndsWith(".json")).ToArray();
            lightshowInfoLeft.val = "Directory:\n"+url;
            string info = "";
            for (int i=0; i<files.Length; i++ )
            {
                lightShowPresets[i] = files[i];
                info += $"{files[i].Split('\\')[1]}\n";
            }
            
            lightshowInfoRight.val = "Files:\n"+info;
            // info.Split('\n')[2].Print();
        }

        private void SaveDefaultLightSettings()
        {
            SaveJSON(GetLightSettingsJSON(), "Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp/DefaultLightSettings.json");
        }

        private void LoadDefaultLightSettings()
        {
            if (!FileManagerSecure.FileExists("Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp/DefaultLightSettings.json"))
                return;
            JSONClass jc = LoadJSON("Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp/DefaultLightSettings.json").AsObject;
            LoadLightSettingsFromJSON(jc);
        }
        
        private void ResetLights()
        {
            lightshowEnabled.val = false;
            isBlending = false;
            headLights.ForEach(x => x.Reset());
        }

        private void ReloadSceneConfig()
        {
            // GetOrAddScenePlugin();
            scenePlugin.CallAction("ReloadSceneConfig");
        }
        
        private JSONClass CreateFakePluginPresetJSON()
        {
            string path = "Custom/Scripts/CheesyFX/FocusOnMe/ScenePlugin/FocusOnMe_ScenePlugin.cs";
            if (Utils.IsInPackage(this))
            {
                var packageUid = Utils.GetPluginPath(this).Split(':')[0] + ":/";
                path = packageUid + path;
            }
            var storables = new JSONArray();
            storables.Add(new JSONClass
            {
                ["id"] = "PluginManager",
                ["plugins"] = new JSONClass
                {
                    ["plugin#0"] = path,
                },
            });
            var jc = new JSONClass
            {
                ["setUnlistedParamsToDefault"] = { AsBool = true },
                ["storables"] = storables,
            };
            return jc;
        }
    }

    public static class JSONStorableParamExt
    {
        public static void Store(this JSONStorableParam jsonParam, JSONClass jc)
        {
            if (jsonParam is JSONStorableBool)
            {
                JSONStorableBool jParam = (JSONStorableBool)jsonParam;
                jc[jParam.name].AsBool = jParam.val;
            }
            else if (jsonParam is JSONStorableFloat)
            {
                JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
                jc[jParam.name].AsFloat = jParam.val;
            }
            else if (jsonParam is JSONStorableString)
            {
                JSONStorableString jParam = (JSONStorableString)jsonParam;
                jc[jParam.name] = jParam.val;
            }
            else if (jsonParam is JSONStorableStringChooser)
            {
                JSONStorableStringChooser jParam = (JSONStorableStringChooser)jsonParam;
                jc[jParam.name] = jParam.val;
            }
            else if (jsonParam is JSONStorableColor)
            {
                JSONStorableColor jParam = (JSONStorableColor)jsonParam;
                JSONArray ja = new JSONArray();
                ja[0].AsFloat = jParam.val.H;
                ja[1].AsFloat = jParam.val.S;
                ja[2].AsFloat = jParam.val.V;
                jc[jParam.name] = ja;
            }
        }
        
        public static void Load(this JSONStorableParam jsonParam, JSONClass jc)
        {
            if (jsonParam is JSONStorableBool)
            {
                JSONStorableBool jParam = (JSONStorableBool)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsBool;
            }
            else if (jsonParam is JSONStorableFloat)
            {
                JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsFloat;
            }
            else if (jsonParam is JSONStorableString)
            {
                JSONStorableString jParam = (JSONStorableString)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name];
            }
            else if (jsonParam is JSONStorableStringChooser)
            {
                JSONStorableStringChooser jParam = (JSONStorableStringChooser)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name];
            }
            else if (jsonParam is JSONStorableColor)
            {
                JSONStorableColor jParam = (JSONStorableColor)jsonParam;
                if (jc.HasKey(jParam.name))
                {
                    HSVColor loadedColor = new HSVColor
                    {
                        H = jc[jParam.name][0].AsFloat,
                        S = jc[jParam.name][1].AsFloat,
                        V = jc[jParam.name][2].AsFloat
                    };
                    jParam.val = loadedColor;
                }
            }
        }
    }
}