using System.Collections;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class FocusOnMe_ScenePlugin : MVRScript
    {
        private const string saveDir = "Saves/PluginData/CheesyFX/FocusOnMe/LightMeUp";
        
        public static JSONClass sceneLightConfig;
        public static JSONClass lightshowConfig;
        private static JSONStorable lightMeUp;

        private JSONStorableAction applySceneLightConfig = new JSONStorableAction("ReloadSceneConfig", ApplySceneLightConfig);
        
        private JSONStorableStringChooser presetChooser0;
        private JSONStorableStringChooser presetChooser1;
        private JSONStorableFloat blendPresets = new JSONStorableFloat("BlendPresets0/1", 0f,0f,1f);
        
        
        private JSONStorableAction blendToRandomPreset = new JSONStorableAction("Trigger blend to random", null);

        private JSONStorableBool lightshowEnabled;
        public JSONStorableBool allowMirroredLR;
        public JSONStorableBool allowMirroredB;
        private JSONStorableUrl tempUrl = new JSONStorableUrl("poo", "", "json", saveDir+"/Presets/");
        private JSONStorableActionPresetFilePath lightShowUrlAction;
        
        private JSONStorableUrl presetURL = new JSONStorableUrl("presetURL", "", "json", saveDir+"/Presets/");
        private JSONStorableActionPresetFilePath applyPresetByUrl;
        
        
        private JSONStorableColor setColor0 = new JSONStorableColor("SetColorLeft", new HSVColor(), SetColor0);

        
        public JSONStorableFloat blendSpeedMean;
        public JSONStorableFloat blendSpeedDelta;
        public JSONStorableFloat delayMean;
        public JSONStorableFloat delayDelta;

        public override void Init()
        {
            applyPresetByUrl = new JSONStorableActionPresetFilePath("ApplyPresetByUrl", ApplyPresetByURL, presetURL);
            lightShowUrlAction = new JSONStorableActionPresetFilePath("Lightshow Dir", null, tempUrl);
            var textfied = CreateTextField(new JSONStorableString("info",
                "This auxiliary plugin was automatically created by 'FocusOnMe.cslist'.\n\n" +
                "It saves and restores the light settings associated with this scene.\n" +
                "Additionally, it provides storables so you can change the settings and behaviour of FocusOnMe via ingame triggers.\n" +
                "You can find it on the atom 'CoreControl'.\n\n" +
                "Do not remove or reload it, otherwise the light data shipped with the scene and the trigger connections are lost."));
            textfied.height = 450f;
            DeferredFindSessionPlugin();
        }

        public void Start()
        {
            ApplySceneLightConfig();
        }
        
        private void BlendPresets(float val)
        {
            lightMeUp.gameObject.SendMessage("BlendPresets", val);
        }
        
        private static void SetColor0(JSONStorableColor colorJson)
        {
            lightMeUp.gameObject.SendMessage("SetColor0", colorJson);
        }

        private void ApplyPresetByURL(string url)
        {
            var json = LoadJSON(url).AsObject;
            lightMeUp.gameObject.SendMessage("LoadLightSettingsFromJSON", json);
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            jc["LightMeUp"]["SceneLightConfig"] = sceneLightConfig;
            jc["LightMeUp"]["LightshowConfig"] = lightshowConfig;
            needsStore = true;
            return jc;
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            StartCoroutine(DeferredRestoreFromJSON(jc));
            
        }

        private IEnumerator DeferredRestoreFromJSON(JSONClass jc)
        {
            if (jc.HasKey("LightMeUp"))
            {
                JSONClass lmu = jc["LightMeUp"].AsObject;
                if(lmu.HasKey("SceneLightConfig")) sceneLightConfig = lmu["SceneLightConfig"].AsObject;
                if(lmu.HasKey("LightshowConfig")) lightshowConfig = lmu["LightshowConfig"].AsObject;
            }
            while (!connected)
            {
                yield return new WaitForEndOfFrame();
            }
            base.LateRestoreFromJSON(jc, true, true, true);
            ApplySceneLightConfig();
        }

        public void FetchSceneLightConfig(JSONClass jc)
        {
            sceneLightConfig = jc;
        }
        
        public void FetchLightshowConfig(JSONClass jc)
        {
            lightshowConfig = jc;
        }

        private static void ApplySceneLightConfig()
        {
            lightMeUp.gameObject.SendMessage("LoadLightSettingsFromJSON", sceneLightConfig);
            lightMeUp.gameObject.SendMessage("LoadLightshowSettingsFromJSON", lightshowConfig);
        }

        private void DeferredFindSessionPlugin()
        {
            StartCoroutine(IEDeferredFindSessionPlugin());
        }

        private IEnumerator IEDeferredFindSessionPlugin()
        {
            yield return new WaitForEndOfFrame();
            if (!FindSessionPlugin())
            {
                SuperController.LogError("FocusOnMe!: FocusOnMe.cslist not found. Install it as a session plugin.");
                yield break;
            }
            ConnectAndRegister();
        }
        
        public bool FindSessionPlugin()
        {
            Transform sessionPlugins = GameObject.Find("SceneAtoms/CoreControl/SessionPluginManagerContainer/CoreControl/SessionPluginManager/Plugins")?.transform;
            if ((object)sessionPlugins == null) return false;
            for (int i=0; i<sessionPlugins.childCount; ++i)
            {
                Transform pluginTransform = sessionPlugins.GetChild(i);
                
                if (pluginTransform.name.EndsWith("CheesyFX.LightMeUp"))
                {
                    lightMeUp = pluginTransform.GetComponent<JSONStorable>();
                    // lightMeUp.SendMessage("UpdatePresetChoices");
                    return true;
                }
            }
            return false;
        }

        private bool connected;
        private bool registered;
        private void ConnectAndRegister()
        {
            Deregister();
            presetChooser0 = lightMeUp.GetStringChooserJSONParam("presetChooser0");
            presetChooser1 = lightMeUp.GetStringChooserJSONParam("presetChooser1");
            blendPresets = lightMeUp.GetFloatJSONParam("Blend presets 0/1");

            blendToRandomPreset.actionCallback = () => lightMeUp.gameObject.SendMessage("InitBlendToNext");
            lightshowEnabled = lightMeUp.GetBoolJSONParam("Lightshow enabled");
            allowMirroredLR = lightMeUp.GetBoolJSONParam("Allow auto mirror L/R");
            allowMirroredB = lightMeUp.GetBoolJSONParam("Allow auto mirror backlight");
            var url = lightMeUp.GetUrlJSONParam("LightShowDirectory");
            lightShowUrlAction.actionCallback = val => url.val = val.Remove(val.LastIndexOf('/'));
            
            blendSpeedMean = lightMeUp.GetFloatJSONParam("BlendSpeedMean");
            blendSpeedDelta = lightMeUp.GetFloatJSONParam("BlendSpeedDelta");
            delayMean = lightMeUp.GetFloatJSONParam("DelayMean");
            delayDelta = lightMeUp.GetFloatJSONParam("DelayDelta");
            
            Register();

            connected = true;
        }

        private void Register()
        {
            RegisterStringChooser(presetChooser0);
            RegisterStringChooser(presetChooser1);
            RegisterFloat(blendPresets);
            
            RegisterAction(applySceneLightConfig);
            RegisterPresetFilePathAction(applyPresetByUrl);
            
            RegisterAction(blendToRandomPreset);
            RegisterBool(lightshowEnabled);
            RegisterBool(allowMirroredLR);
            RegisterBool(allowMirroredB);
            RegisterPresetFilePathAction(lightShowUrlAction);
            RegisterFloat(blendSpeedMean);
            RegisterFloat(blendSpeedDelta);
            RegisterFloat(delayMean);
            RegisterFloat(delayDelta);

            registered = true;
        }

        private void Deregister()
        {
            if (!registered) return;
            DeregisterStringChooser(presetChooser0);
            DeregisterStringChooser(presetChooser1);
            DeregisterFloat(blendPresets);
            
            DeregisterAction(applySceneLightConfig);
            DeregisterPresetFilePathAction(applyPresetByUrl);
            
            DeregisterAction(blendToRandomPreset);
            DeregisterBool(lightshowEnabled);
            DeregisterBool(allowMirroredLR);
            DeregisterBool(allowMirroredB);
            DeregisterPresetFilePathAction(lightShowUrlAction);
            DeregisterFloat(blendSpeedMean);
            DeregisterFloat(blendSpeedDelta);
            DeregisterFloat(delayMean);
            DeregisterFloat(delayDelta);
        }
    }
    
    // public static class Ext1
    // {
    //     public static void Print(this float v)
    //     {
    //         SuperController.LogMessage(v.ToString());
    //     }
    //     
    //     public static void Print(this string v)
    //     {
    //         SuperController.LogMessage(v);
    //     }
    //     
    //     public static void Print(this bool v)
    //     {
    //         SuperController.LogMessage(v.ToString());
    //     }
    //     
    //     public static void NullCheck(this object v)
    //     {
    //         SuperController.LogMessage((v == null).ToString());
    //     }
    // }
    
}