using System;
using System.Collections.Generic;
using System.Linq;
using MacGruber;
using UnityEngine;
using UnityEngine.Serialization;
using Object = UnityEngine.Object;

namespace CheesyFX
{
	public class ViewScan : MVRScript
	{
		// private LightMeUp lightMeUp;
		public DepthAdjust depthAdjust;
		public static bool triggerObjectExists;
		public static JSONStorableFloat width = new JSONStorableFloat("Trigger Width", .1f, SetScale, 0f, 2f, false, true);
		public static JSONStorableFloat height = new JSONStorableFloat("Trigger Height", .1f, SetScale, 0f, 2f, false, true);
		public static JSONStorableFloat heightOffset = new JSONStorableFloat("Trigger Height Offset", 0f, SetScale, -30f, 30f, false, true);
		public static JSONStorableBool debug = new JSONStorableBool("Debug", false);
		public JSONStorableFloat uses = new JSONStorableFloat("uses", 0f, 0f, 5f, true, false);
		public static GameObject triggerObject;
		public static List<GameObject> debugLines = new List<GameObject>();
		public static Collider collider;
		private static float _depth = 10f;
		
		public JSONStorableStringChooser targetChooser = new JSONStorableStringChooser(
			"targetChooser",
			new List<string>{"Persons", "Females", "Males", "Everything"},
			"Persons",
			"Targets");

		public JSONStorableFloat dist = new JSONStorableFloat("targetDistance", 0f, 0f, 10f, false);
		public MyJSONStorableVector3 targetPoint = new MyJSONStorableVector3("targetPoint", new Vector3(), new Vector3(), new Vector3(100f,100f,100f));
		public static float triggerDepth
		{
			set
			{
				if (value < 0)
				{
					_depth = 10f;
					SetTriggerDepth();
					return;
				}
				if (Math.Abs(value - _depth) < .01f) return;
				_depth = value;
				SetTriggerDepth();
			}
		}

		private bool abort;

		private void OnEnable()
		{
			abort = transform.parent.parent.name != "SessionPluginManager";
			uses.setCallbackFunction += SetUses;
		}

		public override void Init()
		{
			if (abort) return;
			// uses.val.ToString().Print();
			CreateSlider(width);
			CreateSlider(height);
			CreateSlider(heightOffset);
			
			CreateToggle(debug);
			CreateFilterablePopup(targetChooser, true);
			
			targetChooser.setCallbackFunction += OnTargetsChanged;
			debug.setCallbackFunction += OnDebugChanged;
			RegisterFloat(uses);
			uses.isStorable = false;
			uses.isRestorable = false;
			RegisterFloat(width);
			RegisterFloat(height);
			RegisterFloat(heightOffset);
			RegisterBool(debug);
			RegisterStringChooser(targetChooser);
			SuperController.singleton.onAtomUIDsWithForceReceiversChangedHandlers += val => OnTargetsChanged(targetChooser.val);
			// OnTargetsChanged(targetChooser.defaultVal);
		}
		
		private void Start()
		{
			if (abort)
			{
				enabledJSON.val = false;
				return;
			}
		}

		private void SetUses(float val)
		{
			if (val == 0)
			{
				OnDestroy();
				return;
			}
			if (!triggerObjectExists) CreateTriggerObject();
			// uses.val.ToString().Print();
		}

		// private void Start()
		// {
		// 	// transform.parent.parent.name.Print();
		// 	bool isSessionPlugin = transform.parent.parent.name == "SessionPluginManager";
		// 	isSessionPlugin.Print();
		// 	MVRScript plugin;
		// 	ViewScanExtensions.FindSessionPlugin("ViewScan", out plugin);
		// 	if(plugin == this) "It's me!".Print();
		// 	else if(plugin != null) "Session intruder!".Print();
		// }
		
		public static Transform GetCameraRig()
		{
			Transform cameraTransform;
			SuperController singleton = SuperController.singleton;
			if (singleton.isOpenVR)
			{
				cameraTransform = singleton.ViveCenterCamera.transform;
			}
			else if (singleton.isOVR)
			{
				cameraTransform = singleton.OVRCenterCamera.transform;
			}
			else
			{
				cameraTransform = singleton.MonitorCenterCamera.transform;
			}

			return cameraTransform;
		}
		
		public GameObject CreateTriggerObject(){
			Vector3 scale = new Vector3(width.val, height.val, _depth);
			Transform cameraTransform = SuperController.singleton.lookCamera.transform;
			triggerObject = GameObject.CreatePrimitive(PrimitiveType.Cube);
			Renderer renderer = triggerObject.GetComponent<Renderer>();
			Material material = new Material(Shader.Find("Transparent/Diffuse"));
            triggerObject.transform.position = cameraTransform.position;
            triggerObject.transform.parent = cameraTransform;
            triggerObject.transform.localScale = scale;
			triggerObject.transform.localEulerAngles = new Vector3(0f, 0f, 0f);
			triggerObject.transform.localPosition = new Vector3(0f, 0f, scale.z / 2);
			collider = triggerObject.GetComponent<Collider>();
			collider.name = "CheesyFX.ViewTrigger";
            collider.enabled = true;
			collider.isTrigger = true;
			renderer.material = material;
			renderer.enabled = false;
			ToggleDebugLines(debug.defaultVal);
			IgnoreCollisionWithTriggers();
			SuperController.singleton.onAtomAddedHandlers += IgnoreCollisionWithTriggers;
			depthAdjust = triggerObject.AddComponent<DepthAdjust>();
			depthAdjust.Init(this);
			triggerObjectExists = true;
			return triggerObject;
		}

		static void IgnoreCollisionWithTriggers(Atom addedAtom = null){
			if(addedAtom == null)
			{
				foreach (Atom atom in SuperController.singleton.GetAtoms())
				{
					foreach (Rigidbody rb in atom.rigidbodies)
					{
						Collider col = rb.GetComponent<Collider>();
						if (col != null && col.isTrigger) Physics.IgnoreCollision(col, collider);
					}
				}
			}
			else
			{
				foreach (Rigidbody rb in addedAtom.rigidbodies)
				{
					Collider col = rb.GetComponent<Collider>();
					if (col != null && col.isTrigger) Physics.IgnoreCollision(col, collider);
				}
			}
		}

		static void SetScale(JSONStorableFloat axis)
		{
			if (axis == heightOffset) triggerObject.transform.localRotation = Quaternion.Euler(heightOffset.val, 0f, 0f);
			{
				Vector3 scale = triggerObject.transform.localScale;
				if (axis == width) scale.x = axis.val;
				if (axis == height) scale.y = axis.val;
				triggerObject.transform.localScale = scale;
			}
			if(debug.val){
				ToggleDebugLines(false);
				ToggleDebugLines(true);
			}
		}

		public static void SetTriggerDepth()
		{
			Vector3 scale = triggerObject.transform.localScale;
			triggerObject.transform.localScale = new Vector3(scale.x, scale.y, _depth);
			triggerObject.transform.localPosition = new Vector3(0f, 0f, _depth / 2);
			if(debug.val){
				ToggleDebugLines(false);
				ToggleDebugLines(true);
			}
		}
		
		static void CreateDebugLine(Vector3 start, Vector3 direction)
		{
			Transform cameraTransform = SuperController.singleton.lookCamera.transform;
			GameObject myLine = new GameObject();
			myLine.transform.position = cameraTransform.position;
			myLine.transform.parent = cameraTransform;
			myLine.transform.localEulerAngles = new Vector3(0f, 0f, 0f);
			LineRenderer lr = myLine.AddComponent<LineRenderer>();
			lr.useWorldSpace = false;
			lr.SetColors(Color.red, Color.yellow);
			lr.SetWidth(0.001f, 0.002f);
			lr.SetPosition(0, start);
			lr.SetPosition(1, start+direction);
			
			// GameObject.Destroy(myLine, duration);
			debugLines.Add(myLine);
		}

		static void ToggleDebugLines(bool enabled){
			if(enabled){
				Vector3 d1 = new Vector3(0f, 0f, _depth);
				List<Vector3> startPositions = new List<Vector3>{
					new Vector3(-width.val/2f, height.val/2f, 0f),
					new Vector3(width.val/2f, height.val/2f, 0f),
					new Vector3(width.val/2f, -height.val/2f, 0f),
					new Vector3(-width.val/2f, -height.val/2f, 0f),
					new Vector3(-width.val/2f, 0f, 0f),
					new Vector3(0f, height.val/2f, 0f),
					new Vector3(width.val/2f, 0f, 0f),
					new Vector3(0f, -height.val/2f, 0f),
				};
				startPositions.ForEach(x => CreateDebugLine(x, d1));
			}
			else{
				debugLines.ForEach(x => Destroy(x));
			}
		}
		
		public void OnTargetsChanged(string val)
		{
			if (depthAdjust == null) return;
			switch (val)
			{
				case "Females":
				{
					depthAdjust.targets = SuperController.singleton.GetAtoms().Where(a =>
						a.type == "Person" &&
						!a.GetComponentInChildren<DAZCharacterSelector>().selectedCharacter.isMale
					).ToList();
					break;
				}
				case "Males":
				{
					depthAdjust.targets = SuperController.singleton.GetAtoms().Where(a =>
						a.type == "Person" &&
						a.GetComponentInChildren<DAZCharacterSelector>().selectedCharacter.isMale
					).ToList();
					break;
				}
				case "Persons":
				{
					depthAdjust.targets = SuperController.singleton.GetAtoms().Where(a =>
						a.type == "Person"
					).ToList();
					break;
				}
				case "Everything":
				{
					depthAdjust.targets = null;
					break;
				}
			}
			// depthAdjust.Rescan();
		}

		public void OnDestroy(){
			Destroy(triggerObject);
			debug.val = false;
			triggerObjectExists = false;
			SuperController.singleton.onAtomAddedHandlers -= IgnoreCollisionWithTriggers;
			SuperController.singleton.onAtomUIDsWithForceReceiversChangedHandlers -= val => OnTargetsChanged(targetChooser.val);
		}

		private void OnDebugChanged(bool val)
		{
			if (val)
			{
				CreateSlider(dist);
				CreateSlider(uses);
			}
			else
			{
				RemoveSlider(dist);
				RemoveSlider(uses);
			}
			ToggleDebugLines(val);
			depthAdjust.debug = val;
		}
	}
	
    

    public class DepthAdjust : MonoBehaviour
    {
	    private ViewScan viewScan;
		private Collider _closest;
		// public float dist;
		private float _minDist = float.MaxValue;
		private Transform _camera;
		public List<Atom> targets = new List<Atom>();
 
		private float maxTriggerDepth = 10f;
		
		public float speed;
		public GameObject debugObj;
		private bool _debug;
 
		public bool debug
		{
			set
			{
				_debug = value;
				if (value)
				{
					CreateDebugObj();
					return;
				}
				Destroy(debugObj);
			}
		}
 
		public void Init(ViewScan viewScan)
		{
			this.viewScan = viewScan;
			_camera = ViewScan.GetCameraRig();
			viewScan.OnTargetsChanged(viewScan.targetChooser.val);
			Rescan();
		}
		
		public void OnTriggerEnter(Collider other)
		{
			// other.name.Print();
			if (targets != null && !targets.Contains(ViewScanExtensions.GetAtom(other))) return;
			float dist = Vector3.Distance(other.ClosestPoint(_camera.position), _camera.position);
			if (dist < _minDist)
			{
				_closest = other;
				_minDist = dist;
				ViewScan.triggerDepth = dist+.1f;
			}
			
		}

		public void OnTriggerExit(Collider other)
		{
			if (other == _closest) Rescan();
		}

		public void Rescan()
		{
			_minDist = float.MaxValue;
			ViewScan.triggerDepth = -1f;
		}
 
		public void Update()
		{
			Vector3 camPos = _camera.position;
			Vector3 closestPoint = _closest.ClosestPoint(camPos);
			viewScan.targetPoint.val = closestPoint;
			viewScan.dist.val = Vector3.Distance(closestPoint, camPos);
			if (_debug) debugObj.transform.position = closestPoint;
		}
		
		public void CreateDebugObj()
		{
			if(debugObj != null) Destroy(debugObj);
			debugObj = GameObject.CreatePrimitive(PrimitiveType.Sphere);
			debugObj.transform.localScale = new Vector3(.01f, .01f, .01f);
			Destroy(debugObj.GetComponent<Collider>());
			var renderer = debugObj.GetComponent<Renderer>();
			var material = renderer.material;
			material.shader = Shader.Find("Battlehub/RTGizmos/Handles");;
			material.color = new Color(0f, 0.89f, 1f);
			debugObj.tag = "debugObj";
		}
    }

    public static class ViewScanExtensions
    {
	    public static Atom GetAtom(this Collider collider)
	    {
		    Atom atom = null;
		    Transform parent = collider.attachedRigidbody.transform.parent;
		    int i = 0;
		    while (atom == null && i < 20)
		    {
			    atom = parent.gameObject.GetComponent<Atom>();
			    parent = parent.parent;
			    i += 1;
		    }

		    return atom;
	    }
	    
	    public static bool FindSessionPlugin(string name, out JSONStorable plugin)
	    {
		    plugin = null;
		    Transform sessionPlugins = GameObject.Find("SceneAtoms/CoreControl/SessionPluginManagerContainer/CoreControl/SessionPluginManager/Plugins")?.transform;
		    if (sessionPlugins == null) return false;
		    for (int i=0; i<sessionPlugins.childCount; ++i)
		    {
			    Transform pluginTransform = sessionPlugins.GetChild(i);
			    if (pluginTransform.name.EndsWith(name))
			    {               
				    plugin = pluginTransform.GetComponent<MVRScript>();
				    return true;
			    }
		    }
		    return false;
	    }

	    public static bool FindScenePlugin(string name, out JSONStorable storable)
	    {
		    storable = null;
		    foreach (var atom in SuperController.singleton.GetAtoms())
		    {
			    var storableId = atom.GetStorableIDs().FirstOrDefault(x => x.EndsWith(name));
			    if (storableId != null)
			    {
				    storable = atom.GetStorableByID(storableId);
				    return true;
			    }
		    }
		    return false;
	    }
    }
}